PROGRAM DX; {Data extraction -- Copyright 1996 Paul Milenkovic}

USES StrBloq;

CONST
    max_col = 128;
    max_row = 16000;

TYPE
    stringP   = ^ASCIIZ_string;
    data_row  = ARRAY [1 .. max_col] OF DOUBLE;
    data_col  = ARRAY [1 .. max_row] OF SINGLE;
    data_colP = ^data_col;

VAR
    errfile: TEXT;

{Parameter control}
    pfile: TEXT;
    pkey, pval: ARRAY [1 .. 64] OF stringP;
    p_used:     ARRAY [1 .. 64] OF BOOLEAN;
    pnum, pidx: INTEGER;

{Parameter values}
    pline, rec, lbl: ASCIIZ_string;
    data_dt, data_scale, time_scale, bad_data: DOUBLE;
    t1, t2, t1_ref, t2_ref, init_data_t: DOUBLE;
    output_t1, output_t2, output_dt, init_output_t: DOUBLE;
    decimals, lbl_occur, icol, fcol, output_n: INTEGER;
    dif_encode, output_dif, time_markers, lbl_message: BOOLEAN;
    ref_initl, ref_final, ref_n, round_dt: BOOLEAN;

{Data blocks}
    data_list: ARRAY [1 .. max_col] OF data_colP;
    deci_list: ARRAY [1 .. max_col] OF INTEGER;
    num_col, num_row: INTEGER;
    mem_limit: BOOLEAN;


PROCEDURE SetPVal;

    VAR i, e: INTEGER;
        pstr: ASCIIZ_string;

BEGIN
    pnum := 0;
    FOR i := 2 TO ParamCount DO BEGIN
        pstr := ParamStr(i);
        e := Pos('=',pstr);
        IF e > 0 THEN BEGIN
            Inc(pnum);
            New(pkey[pnum]);
            New(pval[pnum]);
            pkey[pnum]^  := Concat('#',Copy(pstr,1,Pred(e)),'#');
            pval[pnum]^  := Copy(pstr,Succ(e),Length(pstr)-e);
            p_used[pnum] := FALSE;
        END;
    END;

    ref_initl := FALSE;
    ref_final := FALSE;
    output_n  := 2;
    output_dt := 0;
    ref_n     := TRUE;
    num_col   := 0;
    num_row   := 0;
    mem_limit := FALSE;
END;


PROCEDURE SubPVal(VAR pline: ASCIIZ_string);

    VAR i, p: INTEGER;
        done: BOOLEAN;

BEGIN
    REPEAT
        done := TRUE;
        FOR i := 1 TO pnum DO BEGIN
            p := Pos(pkey[i]^,pline);
            IF p > 0 THEN BEGIN
                done := FALSE;
                Delete(pline,p,Length(pkey[i]^));
                Insert(pval[i]^,pline,p);
                p_used[i] := TRUE;
            END;
        END;
    UNTIL done;
END;


PROCEDURE FindLabelOccur;

    VAR lname: ASCIIZ_string;
        lfile: TEXT;
        lbl_t1, lbl_t2: DOUBLE;
        lbl_val: ASCIIZ_string;
        blank_char: CHAR;
        n: INTEGER;

BEGIN
    IF NOT time_markers THEN BEGIN
        lname := NewExt(rec,'LBL');
        IF FileExists(FileInPath(lname)) THEN
            lname := FileInPath(lname);
        Assign(lfile,lname);
        {$I-} Reset(lfile) {$I+};
        IF IOResult = 0 THEN BEGIN
            n := 0;
            WHILE (n < lbl_occur) AND NOT EOF(lfile) DO BEGIN
                Readln(lfile,lbl_t1,lbl_t2,blank_char,lbl_val);
                IF lbl_val = lbl THEN BEGIN
                    Inc(n);
                    IF n = lbl_occur THEN BEGIN
                        time_markers := TRUE;
                        t1 := lbl_t1;
                        t2 := lbl_t2;
                        t1_ref := t1;
                        t2_ref := t2;
                    END;
                END;
            END;
            IF (n < lbl_occur) AND (NOT lbl_message) THEN BEGIN
                Writeln(
                    errfile,
                    'Occurence ',lbl_occur:0,' of label ',lbl,
                    ' in ',lname,' not found'
                );
                lbl_message := TRUE;
            END;
            Close(lfile);
        END ELSE IF NOT lbl_message THEN BEGIN
            Writeln(errfile,'Label file ',lname,' not found');
            lbl_message := TRUE;
        END;
    END;
END;


PROCEDURE OutputDataCols;

    VAR i, j: INTEGER;

BEGIN
    FOR i := 1 TO num_row DO BEGIN
        FOR j := 1 TO num_col DO BEGIN
            Write(data_list[j]^[i]:0:deci_list[j]);
            IF j < num_col THEN
                Write(#9);
        END;
        Writeln;
    END;
    IF output_dif AND (num_row > 1) THEN BEGIN
        FOR j := 1 TO num_col DO BEGIN
            Write(
            data_list[j]^[num_row]-data_list[j]^[num_row-1]:0:deci_list[j]
            );
            IF j < num_col THEN
                Write(#9);
        END;
        Writeln;
    END;
    FOR j := num_col DOWNTO 1 DO
        FreeMem(data_list[j],num_row*Sizeof(SINGLE));
    num_col   := 0;
    num_row   := 0;
    mem_limit := FALSE;
END;


PROCEDURE SetTimeCol;

    VAR i: INTEGER;

BEGIN
    IF (num_row > 0) AND (num_row <> output_n) THEN
        OutputDataCols;
    IF num_col = 0 THEN
        num_row := output_n;
    IF num_row > 0 THEN BEGIN
        IF (num_col < max_col) AND (num_row <= MaxAvail DIV Sizeof(SINGLE))
        THEN BEGIN
            Inc(num_col);
            GetMem(data_list[num_col],num_row*(Sizeof(SINGLE)));
            deci_list[num_col] := 3;
            FOR i := 1 TO num_row DO
                data_list[num_col]^[i] := init_output_t+(i-1)*output_dt;
        END ELSE BEGIN
            IF NOT mem_limit THEN
                Writeln(
                    errfile,
                    'Exceeding memory limit to go beyond ',
                    num_row:0,' rows and ',num_col,' cols'
                );
            mem_limit := TRUE;
        END;
    END;
END;


PROCEDURE SetDataCol;

    VAR i, j: INTEGER;
        data_t0, data_t1, t, dval: DOUBLE;
        cols_t0, cols_t1: data_row;
        ncol: INTEGER;
        dfile: TEXT;
        data_file: BOOLEAN;

BEGIN
    IF (num_row > 0) AND (num_row <> output_n) THEN
        OutputDataCols;
    IF num_col = 0 THEN
        num_row := output_n;
    ncol := Succ(fcol - icol);
    IF num_row > 0 THEN BEGIN
        j := 0;
        WHILE (j < ncol) AND
              (num_col + j < max_col) AND
              (num_row <= MaxAvail div Sizeof(SINGLE))
        DO BEGIN
            Inc(j);
            GetMem(data_list[num_col+j],num_row*(Sizeof(SINGLE)));
            deci_list[num_col+j] := decimals;
        END;
        IF j < ncol THEN BEGIN
            IF NOT mem_limit THEN
                Writeln(
                    errfile,
                    'Exceeding memory limit to go beyond ',
                    num_row:0,' rows and ',num_col+j,' cols'
                );
            mem_limit := TRUE;
            ncol := j;
        END;
        IF ncol > 0 THEN BEGIN
            Assign(dfile,rec);
            {$I-} Reset(dfile) {$I+};
            data_file := IOResult = 0;
            IF NOT data_file THEN
                Writeln(errfile,'Data record file ',rec,' not found');
            data_t1 := -1e-30;
            FOR j := icol TO Pred(icol+ncol) DO
                cols_t1[j] := bad_data;
            FOR i := 1 TO num_row DO BEGIN
                t := init_data_t + (i - 1)*output_dt;
                WHILE (data_t1 < t) AND data_file DO BEGIN
                    IF EOF(dfile) THEN BEGIN
                        Close(dfile);
                        data_file := FALSE;
                    END ELSE BEGIN
                        data_t0 := data_t1;
                        FOR j := icol TO Pred(icol+ncol) DO
                            cols_t0[j] := cols_t1[j];
                        IF NOT dif_encode THEN BEGIN
                            Read(dfile,data_t1);
                            data_t1 := data_t1 * time_scale;
                        END;
                        FOR j := 1 TO Pred(icol+ncol) DO BEGIN
                            Read(dfile,dval);
                            IF j >= icol THEN BEGIN
                                IF (NOT dif_encode) OR (data_t1 < 0) THEN
                                    cols_t1[j] := dval
                                ELSE
                                    cols_t1[j] := cols_t1[j] + dval;
                            END;
                        END;
                        Readln(dfile);
                        IF dif_encode THEN BEGIN
                            IF data_t1 < 0 THEN
                                data_t1 := 0
                            ELSE
                                data_t1 := data_t1 + data_dt;
                        END;
                    END;
                END;
                FOR j := 1 TO ncol DO BEGIN
                    IF data_file THEN BEGIN
                        IF (cols_t0[Pred(icol+j)] = bad_data) OR
                           (cols_t1[Pred(icol+j)] = bad_data)
                        THEN
                            data_list[num_col+j]^[i] := data_scale*bad_data
                        ELSE
                        IF data_t0 = data_t1 THEN
                            data_list[num_col+j]^[i] :=
                            data_scale*cols_t1[Pred(icol+j)]
                        ELSE
                            data_list[num_col+j]^[i] := data_scale*(
                                cols_t0[Pred(icol+j)]*(data_t1 - t) +
                                cols_t1[Pred(icol+j)]*(t - data_t0)
                            )/(data_t1 - data_t0);
                    END ELSE
                        data_list[num_col+j]^[i] := data_scale*bad_data;
                END;
            END;
            IF data_file THEN Close(dfile);
            Inc(num_col,ncol);
        END;
    END;
END;


PROCEDURE SetOutputTimes;

    VAR n: LONGINT;

BEGIN
    IF time_markers THEN BEGIN
        IF (NOT ref_n) AND (output_n = 0) THEN
            OutputDataCols;
        IF (NOT ref_n) AND (output_n = 0) AND (output_dt > 0) THEN BEGIN
            IF round_dt THEN
                n := Round((t2 - t1)/output_dt)
            ELSE
                n := 2 + Trunc((t2 - t1)/output_dt);
        END ELSE
            n := output_n;
        IF n > max_row THEN BEGIN
            output_n := max_row;
            Writeln(errfile,'Time samples limited to max of ',output_n:0);
        END ELSE
            output_n := n;
        IF ref_n OR round_dt THEN BEGIN
            init_data_t := t1;
            IF output_n > 1 THEN
                output_dt := (t2 - t1)/Pred(output_n)
            ELSE
                output_dt := 0;
            IF ref_initl THEN
                init_output_t := output_t1
            ELSE
            IF ref_final THEN
                init_output_t := output_t2 + t1 - t2
            ELSE
                init_output_t := t1;
        END ELSE BEGIN
            IF ref_initl THEN BEGIN
                init_data_t   := t2 - Pred(output_n)*output_dt;
                init_output_t := init_data_t + output_t1 - t1;
            END ELSE
            IF ref_final THEN BEGIN
                init_data_t   := t1;
                init_output_t := init_data_t + output_t2 - t2;
            END ELSE BEGIN
                init_data_t   := Trunc(t1/output_dt)*output_dt;
                init_output_t := init_data_t;
            END;
        END;
    END ELSE
        output_n := 0;
END;


PROCEDURE ExecuteCommand(VAR pline: ASCIIZ_string);

    VAR ext: ASCIIZ_string;
        wt1, wt2, wt1_t2, wt2_t1, t: DOUBLE;

BEGIN
    ext := '';

    IF StringKeyword(rec,pline,'rec=') THEN BEGIN
        ext := PathExt(rec);
        time_markers := FALSE;
        lbl_message  := FALSE;
    END ELSE
    IF StringKeyword(ext,pline,'ext=') THEN BEGIN
        rec := NewExt(rec,ext);
        IF FileExists(FileInPath(rec)) THEN
            rec := FileInPath(rec);
    END ELSE
    IF DoubleKeyword(data_dt,pline,'data_dt=') THEN BEGIN
    END ELSE
    IF IntegerKeyword(decimals,pline,'dec=') THEN BEGIN
    END ELSE
    IF DoubleKeyword(data_scale,pline,'scale=') THEN BEGIN
    END ELSE
    IF DoubleKeyword(bad_data,pline,'bad_data=') THEN BEGIN
    END ELSE
    IF StringKeyword(lbl,pline,'lbl=') THEN BEGIN
        lbl_occur := 1;
        time_markers := FALSE;
        lbl_message  := FALSE;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF IntegerKeyword(lbl_occur,pline,'n=') THEN BEGIN
    END ELSE
    IF DoubleKeyword(t1,pline,'t1=') THEN BEGIN
        t1_ref := t1;
        IF NOT time_markers THEN BEGIN
            time_markers := TRUE;
            t2 := t1;
            t2_ref := t2;
        END;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(t2,pline,'t2=') THEN BEGIN
        t2_ref := t2;
        IF NOT time_markers THEN BEGIN
            time_markers := TRUE;
            t1 := t2;
            t1_ref := t1;
        END;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(wt1,pline,'wt1=') THEN BEGIN
        FindLabelOccur;
        t1 := t1_ref - wt1;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(wt2,pline,'wt2=') THEN BEGIN
        FindLabelOccur;
        t2 := t2_ref + wt2;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(wt1_t2,pline,'wt1_t2=') THEN BEGIN
        FindLabelOccur;
        t1 := t2_ref - wt1_t2;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(wt2_t1,pline,'wt2_t1=') THEN BEGIN
        FindLabelOccur;
        t2 := t1_ref + wt2_t1;
        IF NOT ref_n THEN
            output_n := 0;
    END ELSE
    IF DoubleKeyword(output_t1,pline,'output_t1=') THEN BEGIN
        ref_initl := TRUE;
    END ELSE
    IF DoubleKeyword(output_t2,pline,'output_t2=') THEN BEGIN
        ref_final := TRUE;
    END ELSE
    IF DoubleKeyword(output_dt,pline,'output_dt=') THEN BEGIN
        round_dt := FALSE;
        output_n := 0;
        ref_n := FALSE;
        output_dif := FALSE;
    END ELSE
    IF IntegerKeyword(output_n,pline,'output_n=') THEN BEGIN
        IF NOT ref_n THEN
            OutputDataCols;
        ref_n := TRUE;
        output_dif := FALSE;
    END ELSE
    IF IntegerKeyword(icol,pline,'col=') THEN BEGIN
        fcol := icol;
    END ELSE
    IF IntegerKeyword(icol,pline,'icol=') THEN BEGIN
    END ELSE
    IF IntegerKeyword(fcol,pline,'fcol=') THEN BEGIN
    END ELSE
    IF pline = 'dif_encode' THEN BEGIN
        dif_encode := TRUE;
    END ELSE
    IF pline = 'round_dt' THEN BEGIN
        round_dt := TRUE;
    END ELSE
    IF pline = 'output_endpoints' THEN BEGIN
        FindLabelOccur;
        IF time_markers THEN BEGIN
            output_dif := TRUE;
            output_n   := 2;
            IF NOT ref_n THEN
                OutputDataCols;
            ref_n := TRUE;
        END ELSE BEGIN
            output_dif := FALSE;
            output_n   := 0;
        END;
    END ELSE
    IF pline = 'write_time' THEN BEGIN
        FindLabelOccur;
        SetOutputTimes;
        SetTimeCol;
    END ELSE
    IF pline = 'new_row' THEN BEGIN
        OutputDataCols;
    END ELSE
    IF pline = 'write_data' THEN BEGIN
        FindLabelOccur;
        SetOutputTimes;
        SetDataCol;
    END ELSE
    IF pline = 'write_rec' THEN BEGIN
        OutputDataCols;
        Writeln(rec);
    END ELSE
    IF pline = 'write_lbl' THEN BEGIN
        FindLabelOccur;
        OutputDataCols;
        Writeln(t1:10:3,t2:10:3,' ',lbl);
    END ELSE
        Writeln(errfile,'Command ',pline,' not recognized');

    IF ext <> '' THEN BEGIN
        MakeUpCase(ext);
        bad_data   := 0;
        dif_encode := FALSE;
        decimals   := 3;
        data_scale := 1.0;
        time_scale := 1.0;
        IF ext = 'XYD' THEN BEGIN
            dif_encode := TRUE;
            data_dt    := 6.866;
            data_scale := 0.001;
            bad_data   := 1000000.0;
        END ELSE
        IF ext = 'TXY' THEN BEGIN
            time_scale := 0.001;
            bad_data   := 1000000.0;
        END;
    END;
END;


BEGIN
    Assign(errfile,'CON');
    Rewrite(errfile);
    IF ParamCount > 0 THEN BEGIN
        Assign(pfile,ParamStr(1));
        {$I-} Reset(pfile) {$I+};
        IF IOResult = 0 THEN BEGIN
            Assign(output,'');
            Rewrite(output);
            SetPVal;
            WHILE NOT EOF(pfile) DO BEGIN
                Readln(pfile,pline);
                SubPVal(pline);
                ExecuteCommand(pline);
            END;
            OutputDataCols;
            Close(pfile);
            FOR pidx := 1 TO pnum DO IF NOT p_used[pidx] THEN
                Writeln(errfile,'Parameter ',pkey[pidx]^,' not used');
        END ELSE
            Writeln(errfile,'Parameter file ',ParamStr(1),' not found');
    END ELSE
        Writeln(errfile,'Usage: dx pfile p1=p1val . . . pn=pnval >dfile');
END.